import { defineComponent, h, ref, watch } from 'vue'

import { CFormInput, CFormSelect, CInputGroup, CInputGroupText } from '../form/'
import { CFormControlWrapper } from './../form/CFormControlWrapper'
import { CPicker } from '../picker'
import { CTimePickerRollCol } from './CTimePickerRollCol'

import {
  convert12hTo24h,
  convertTimeToDate,
  getAmPm,
  getListOfHours,
  getListOfMinutes,
  getListOfSeconds,
  getSelectedHour,
  getSelectedMinutes,
  getSelectedSeconds,
  isAmPm,
  isValidTime,
} from '../../utils/time'

import { Color } from '../props'

const CTimePicker = defineComponent({
  name: 'CTimePicker',
  props: {
    ...CPicker.props,
    /**
     * Set if the component should use the 12/24 hour format. If `true` forces the interface to a 12-hour format. If `false` forces the interface into a 24-hour format. If `auto` the current locale will determine the 12 or 24-hour interface by default locales.
     *
     * @since 4.7.0
     */
    ampm: {
      type: [Boolean, String],
      default: 'auto',
      validator: (value: boolean | string) => {
        if (typeof value == 'string') {
          return ['auto'].includes(value)
        }
        if (typeof value == 'boolean') {
          return true
        }
        return false
      },
    },
    /**
     * Toggle visibility or set the content of cancel button.
     */
    cancelButton: {
      type: [Boolean, String],
      default: 'Cancel',
    },
    /**
     * Sets the color context of the cancel button to one of CoreUI’s themed colors.
     *
     * @values 'primary', 'secondary', 'success', 'danger', 'warning', 'info', 'dark', 'light'
     */
    cancelButtonColor: {
      ...Color,
      default: 'primary',
    },
    /**
     * Size the cancel button small or large.
     *
     * @values 'sm', 'lg'
     */
    cancelButtonSize: {
      type: String,
      default: 'sm',
      validator: (value: string) => {
        return ['sm', 'lg'].includes(value)
      },
    },
    /**
     * Set the cancel button variant to an outlined button or a ghost button.
     *
     * @values 'ghost', 'outline'
     */
    cancelButtonVariant: {
      type: String,
      default: 'ghost',
      validator: (value: string) => {
        return ['ghost', 'outline'].includes(value)
      },
    },
    /**
     * Toggle visibility of the cleaner button.
     */
    cleaner: {
      type: Boolean,
      default: true,
    },
    /**
     * Toggle visibility or set the content of confirm button.
     */
    confirmButton: {
      type: [Boolean, String],
      default: 'OK',
    },
    /**
     * Sets the color context of the confirm button to one of CoreUI’s themed colors.
     *
     * @values 'primary', 'secondary', 'success', 'danger', 'warning', 'info', 'dark', 'light'
     */
    confirmButtonColor: {
      ...Color,
      default: 'primary',
    },
    /**
     * Size the confirm button small or large.
     *
     * @values 'sm', 'lg'
     */
    confirmButtonSize: {
      type: String,
      default: 'sm',
      validator: (value: string) => {
        return ['sm', 'lg'].includes(value)
      },
    },
    /**
     * Set the confirm button variant to an outlined button or a ghost button.
     *
     * @values 'ghost', 'outline'
     */
    confirmButtonVariant: {
      type: String,
      validator: (value: string) => {
        return ['ghost', 'outline'].includes(value)
      },
    },
    /**
     * Provide valuable, actionable feedback.
     *
     * @since 4.6.0
     */
    feedback: {
      type: String,
    },
    /**
     * Provide valuable, actionable feedback.
     *
     * @since 4.6.0
     */
    feedbackInvalid: {
      type: String,
    },
    /**
     * Provide valuable, actionable invalid feedback when using standard HTML form validation which applied two CSS pseudo-classes, `:invalid` and `:valid`.
     *
     * @since 4.6.0
     */
    feedbackValid: {
      type: String,
    },
    /**
     * The id global attribute defines an identifier (ID) that must be unique in the whole document.
     */
    id: String,
    /**
     * Toggle visibility or set the content of the input indicator.
     */
    indicator: {
      type: Boolean,
      default: true,
    },
    /**
     * Toggle the readonly state for the component.
     */
    inputReadOnly: Boolean,
    /**
     * Set component validation state to invalid.
     *
     * @since 4.6.0
     */
    invalid: Boolean,
    /**
     * Add a caption for a component.
     *
     * @since 4.6.0
     */
    label: {
      type: String,
    },
    /**
     * Sets the default locale for components. If not set, it is inherited from the navigator.language.
     */
    locale: {
      type: String,
      defalut: 'default',
    },
    /**
     * Specifies a short hint that is visible in the input.
     */
    placeholder: {
      type: String,
      default: 'Select time',
    },
    /**
     * Show seconds.
     *
     * @since 4.7.0
     */
    seconds: {
      type: Boolean,
      default: true,
    },
    /**
     * Size the component small or large.
     *
     * @values 'sm', 'lg'
     */
    size: {
      type: String,
      default: undefined,
      validator: (value: string) => {
        return ['sm', 'lg'].includes(value)
      },
    },
    /**
     * Add helper text to the component.
     *
     * @since 4.6.0
     */
    text: {
      type: String,
    },
    /**
     * Initial selected time.
     */
    time: {
      type: [Date, String],
    },
    /**
     * Display validation feedback in a styled tooltip.
     *
     * @since 4.6.0
     */
    tooltipFeedback: Boolean,
    /**
     * Set component validation state to valid.
     *
     * @since 4.6.0
     */
    valid: Boolean,
    /**
     * Set the time picker variant to a roll or select.
     *
     * @values 'roll', 'select'
     */
    variant: {
      type: String,
      default: 'roll',
      validator: (value: string) => {
        return ['roll', 'select'].includes(value)
      },
    },
    /**
     * Toggle the visibility of the component.
     */
    visible: Boolean,
  },
  emits: [
    /**
     * Callback fired when the time changed.
     */
    'change',
    /**
     * Callback fired when the component requests to be hidden.
     */
    'hide',
    /**
     * Callback fired when the component requests to be shown.
     */
    'show',
    /**
     * Callback fired when the time changed.
     *
     * @since 4.7.0
     */
    'update:time',
  ],
  setup(props, { emit, attrs, slots }) {
    const visible = ref(props.visible)
    const date = ref<Date | null>(convertTimeToDate(props.time))
    const initialDate = ref<Date | null>(null)
    const ampm = ref<'am' | 'pm'>(date.value ? getAmPm(new Date(date.value), props.locale) : 'am')

    watch(
      () => props.time,
      () => {
        date.value = convertTimeToDate(props.time)
      },
    )

    watch(date, () => {
      if (date.value) {
        ampm.value = getAmPm(new Date(date.value), props.locale)
      }
    })

    const handleClear = (event: Event) => {
      event.stopPropagation()
      date.value = null
      emit('change', null)
      emit('update:time', null)
    }

    const handleTimeChange = (set: 'hours' | 'minutes' | 'seconds' | 'toggle', value: string) => {
      const _date = date.value || new Date('1970-01-01')

      if (set === 'toggle') {
        if (value === 'am') {
          _date.setHours(_date.getHours() - 12)
        }
        if (value === 'pm') {
          _date.setHours(_date.getHours() + 12)
        }
      }

      if (set === 'hours') {
        if ((props.ampm === 'auto' && isAmPm(props.locale)) || props.ampm === true) {
          _date.setHours(convert12hTo24h(ampm.value, parseInt(value)))
        } else {
          _date.setHours(parseInt(value))
        }
      }

      if (set === 'minutes') {
        _date.setMinutes(parseInt(value))
      }

      if (set === 'seconds') {
        _date.setSeconds(parseInt(value))
      }

      date.value = new Date(_date)
      emit('change', _date.toTimeString(), _date.toLocaleTimeString(props.locale), _date)
      emit('update:time', _date.toLocaleTimeString(props.locale))
    }

    const InputGroup = () =>
      h(CInputGroup, { class: 'picker-input-group', size: props.size }, () => [
        h(CFormInput, {
          autocomplete: 'off',
          disabled: props.disabled,
          onInput: (event) => {
            if (isValidTime(event.target.value)) {
              date.value = convertTimeToDate(event.target.value)
            }
          },
          placeholder: props.placeholder,
          readonly: props.inputReadOnly,
          value: date.value
            ? date.value.toLocaleTimeString(props.locale, {
                hour12:
                  (props.ampm === 'auto' && isAmPm(props.locale)) ||
                  (props.ampm === 'boolean' && props.ampm),
                ...(!props.seconds && { timeStyle: 'short' }),
              })
            : '',
        }),
        (props.indicator || props.cleaner) &&
          h(CInputGroupText, {}, () => [
            props.indicator &&
              h(
                'span',
                {
                  class: 'picker-input-group-indicator',
                },
                slots.indicator
                  ? slots.indicator()
                  : h('span', { class: 'picker-input-group-icon time-picker-input-icon' }),
              ),
            props.cleaner &&
              h(
                'span',
                {
                  class: 'picker-input-group-cleaner',
                  onClick: (event: Event) => handleClear(event),
                  role: 'button',
                },
                slots.cleaner
                  ? slots.cleaner()
                  : h('span', { class: 'picker-input-group-icon time-picker-cleaner-icon' }),
              ),
          ]),
      ])

    const TimePickerSelect = () => [
      h('span', { class: 'time-picker-inline-icon' }),
      h(CFormSelect, {
        disabled: props.disabled,
        options: getListOfHours(props.locale).map((option) => {
          return {
            value: option.value.toString(),
            label: option.label,
          }
        }),
        onChange: (event) => handleTimeChange('hours', event.target.value),
        ...(date.value && { value: getSelectedHour(date.value, props.locale) }),
      }),
      ':',
      // @ts-expect-error the getListOfMinutes function returns corect type
      h(CFormSelect, {
        disabled: props.disabled,
        options: getListOfMinutes(props.locale, true),
        onChange: (event: Event) =>
          handleTimeChange('minutes', (event.target as HTMLSelectElement).value),
        ...(date.value && { value: getSelectedMinutes(date.value) }),
      }),
      props.seconds && ':',
      props.seconds &&
        // @ts-expect-error the getListOfMinutes function returns corect type
        h(CFormSelect, {
          disabled: props.disabled,
          options: getListOfSeconds(props.locale, true),
          onChange: (event: Event) =>
            handleTimeChange('seconds', (event.target as HTMLSelectElement).value),
          ...(date.value && { value: getSelectedSeconds(date.value) }),
        }),
      isAmPm(props.locale) &&
        h(CFormSelect, {
          disabled: props.disabled,
          options: [
            { value: 'am', label: 'AM' },
            { value: 'pm', label: 'PM' },
          ],
          onChange: (event) => handleTimeChange('toggle', event.target.value),
          value: ampm.value,
        }),
    ]

    const TimePickerRoll = () => [
      h(CTimePickerRollCol, {
        elements: getListOfHours(props.locale, props.ampm),
        onClick: (index: number) => handleTimeChange('hours', index.toString()),
        selected: getSelectedHour(date.value, props.locale, props.ampm),
      }),
      h(CTimePickerRollCol, {
        elements: getListOfMinutes(props.locale),
        onClick: (index: number) => handleTimeChange('minutes', index.toString()),
        selected: getSelectedMinutes(date.value),
      }),
      props.seconds &&
        h(CTimePickerRollCol, {
          elements: getListOfSeconds(props.locale),
          onClick: (index: number) => handleTimeChange('seconds', index.toString()),
          selected: getSelectedSeconds(date.value),
        }),
      ((props.ampm === 'auto' && isAmPm(props.locale)) || props.ampm === true) &&
        h(CTimePickerRollCol, {
          elements: [
            { value: 'am', label: 'AM' },
            { value: 'pm', label: 'PM' },
          ],
          onClick: (value: string) => handleTimeChange('toggle', value),
          selected: ampm.value,
        }),
    ]

    return () =>
      h(
        CFormControlWrapper,
        {
          describedby: attrs['aria-describedby'],
          feedback: props.feedback,
          feedbackInvalid: props.feedbackInvalid,
          feedbackValid: props.feedbackValid,
          id: props.id,
          invalid: props.invalid,
          label: props.label,
          text: props.text,
          tooltipFeedback: props.tooltipFeedback,
          valid: props.valid,
        },
        {
          default: () =>
            h(
              CPicker,
              {
                cancelButton: props.cancelButton,
                cancelButtonColor: props.cancelButtonColor,
                cancelButtonSize: props.cancelButtonSize,
                cancelButtonVariant: props.cancelButtonVariant,
                class: ['time-picker', { 'is-invalid': props.invalid, 'is-valid': props.valid }],
                confirmButton: props.confirmButton,
                confirmButtonColor: props.confirmButtonColor,
                confirmButtonSize: props.confirmButtonSize,
                confirmButtonVariant: props.confirmButtonVariant,
                container: props.container,
                disabled: props.disabled,
                footer: true,
                onCancel: () => {
                  if (initialDate.value) {
                    date.value = new Date(initialDate.value)
                  }
                  visible.value = false
                },
                onHide: () => {
                  visible.value = false
                  emit('hide')
                },
                onShow: () => {
                  if (date.value) {
                    initialDate.value = new Date(date.value)
                  }

                  visible.value = true
                  emit('show')
                },
                visible: visible.value,
              },
              {
                ...(slots.cancelButton && {
                  cancelButton: () => slots.cancelButton && slots.cancelButton(),
                }),
                ...(slots.confirmButton && {
                  confirmButton: () => slots.confirmButton && slots.confirmButton(),
                }),
                toggler: () => InputGroup(),
                default: () =>
                  h(
                    'div',
                    {
                      class: [
                        'time-picker-body',
                        {
                          ['time-picker-roll']: props.variant === 'roll',
                        },
                      ],
                    },
                    props.variant === 'select' ? TimePickerSelect() : TimePickerRoll(),
                  ),
              },
            ),
        },
      )
  },
})
export { CTimePicker }
